(***Type checking algorithm for session types in the pi calculus. See [3,4] in the
References at the end of the page. For an industrial perspective a/o a gentle
introduction to the subject see the interview with Steve Ross-Talbot [5]. 
Presented at the Workshop on Automated Specification and Verification of Web Systems
(WWV), DisCoTec 2011 [1]. Implements the asbtract type system in CONCUR 2010 by Giunti and
Vasconcelos [2]. Ver 1.2,July 2011, Creative Commons.
Contacts: Marco Giunti, INRIA & LIX, Ecole Polytechnique, France
Web: http://www.lix.polytechnique.fr/~marco.giunti
***)

(***We consider input/output session types qualified as linear or unrestricted. A linear
type is used exactly once, possibly multiple times (because of type evolution).
A session end point  In (a,c) describes the input of an argument of type a and then
the evolution to type c. A end point type Out (a,c) describes the output of an arguments
of type a and then the evolution as c. The session end point End describes the end of the
session. Channel types have the form Channel (s1,s2) where each s_i is a session end point
or the special type Void, which is used in the type checking algorithm to forbid the use
of a consumed linear end point. 
***) 


datatype qualifier  = Lin | Un;
datatype  preType = End 
                    | In  of  sessionType *  endpointType  
                    | Out of  sessionType *  endpointType
and  endpointType = Qualified of qualifier *  preType 
                    | Void
and  sessionType  = Channel of  endpointType *  endpointType 
                    | EndPoint of  endpointType
                    | Base;
type var           = string; 
type context       = (var * sessionType) list;  
 

(*****************************Pi calculus processes***************************) 
(***Polymorphic implementation of pi calculus process. Channels carry channels 
and values of base type ***)

datatype 'a process = Zero
                    | Replication of 'a process 
                    | Parallel of 'a process * 'a process 
                    | Input of string * string * 'a process 
                    | Output of string * 'a value * 'a process 
                    | New of string *  sessionType * 'a process
and      'a value   = Value of 'a | Var of string; 


(******************Exception declarations************************************)

exception typeFail;
exception contextFail;
exception unSafe;
exception notConsumed;
exception checkVarFail;   
exception checkFail;   

(*****************************Functions over session types*********************)


(*** Setting a end point type to void ***)
fun setVoid (EndPoint s) = Void
 |  setVoid (_) = raise typeFail;



(* Casting of two end point types to a channel type *)
fun upCast (EndPoint e, EndPoint f)= Channel (e,f)
 |  upCast (_,_) = raise typeFail;

(* Casting of a session type to an end point type *)
fun downCast ( EndPoint e)= e
 |  downCast (_) = raise typeFail

(*** Defining the dual end point types. Used in fun safe ***)
fun dual ( Qualified  (p,End), Qualified  (q,End)) = 
    p=q
 |  dual (Qualified  (p,In (a,c)), Qualified  (q,Out (b,d)) ) =
     p=q andalso a=b andalso dual (c,d)
 |  dual (Qualified  (p,Out (a,c)), Qualified  (q,In (b,d)) ) =
     p=q andalso a=b andalso dual (c,d)
 |  dual(_,_) = raise unSafe;
     
 
(*** Defining the class of safe types. Used in fun safeContext  and in the restriction
case in fun check ***)
fun safe (Base) = Base
 |  safe (EndPoint (Qualified  (q,End))) = EndPoint (Qualified  (q,End))
 |  safe (EndPoint (Qualified  (q,In (a,c)))) =
      let val t = downCast (safe ( EndPoint (c) ) ) 
       in EndPoint (Qualified  (q,In (safe(a),t) ) )
      end
 |  safe (EndPoint (Qualified  (q,Out (a,c)))) =
      let val t = downCast (safe ( EndPoint (c) ) ) 
       in EndPoint (Qualified  (q,Out (safe(a),t) ) )
      end
 |  safe (Channel (e,g)) = 
      let val s = safe (EndPoint e) in
       let val t = safe (EndPoint g) in
        if dual (downCast (s),downCast(t) ) then
         upCast (  s ,  t )
        else
         raise unSafe
       end
      end
 | safe(_) = raise unSafe;
  

(* Checking if a type is unrestricted or void. In case, returns the session type
received in input. Uses fun upCast. Used in fun consumed. *)
fun unRestricted (EndPoint Void) = EndPoint Void
 |  unRestricted (EndPoint (Qualified (Un,p))) = EndPoint (Qualified (Un,p))  
 |  unRestricted (Channel  (e,f)) =  
     let val s = unRestricted (EndPoint e) 
      in let val t= unRestricted (EndPoint f)
       in upCast (s,t)
      end
     end
 | unRestricted(_) = raise notConsumed;
     
     
(*Composing two types. The return type to is set to void if the two types are 
equal and linear. Used in fun checkVar.*)

fun compType (Base,Base) = Base
 |  compType (EndPoint (Void),EndPoint (Void)) = EndPoint (Void) 
 |  compType (EndPoint (Qualified (Un,p)),EndPoint (Qualified (Un,r))) =
     if p=r then
      EndPoint (Qualified (Un,p))
     else
      raise typeFail
 |  compType (EndPoint (Qualified (Lin,p)),EndPoint (Qualified (Lin,r))) =
     if p=r then
      EndPoint (Void)
     else
      raise typeFail
 |  compType (Channel  ( Qualified (Un,p), Qualified (Un,s) ),  
               Channel  ( Qualified (Un,r), Qualified (Un,t) )) =
     if p=r andalso s=t then 
          Channel  ( Qualified (Un,p), Qualified (Un,s))  
     else
      if p=t andalso s=r then
          Channel  ( Qualified (Un,p), Qualified (Un,s)) 
      else
       raise typeFail
 |  compType (Channel  ( Qualified (Lin,p), Qualified (Lin,s) ),  
               Channel  ( Qualified (Lin,r), Qualified (Lin,t) )) =
     if p=r andalso s=t then 
          Channel  ( Void, Void ) 
     else
      if p=t andalso s=r then
          Channel  ( Void, Void )
      else
       raise typeFail
 | compType (Channel   ( Qualified (Un,p),  e), EndPoint  ( Qualified (Un,r)))=
     if p=r   then 
       (Channel   ( Qualified (Un,p),  e))  
     else 
       let val s = compType (EndPoint (e), EndPoint ( Qualified (Un,r)))
        in let val t=downCast(s) 
         in Channel ( Qualified (Un,p),t )
        end
       end
 | compType (Channel   ( e,  Qualified (Un,p)), EndPoint  ( Qualified (Un,r)))=
     if p=r   then 
       Channel ( e,Qualified (Un,p) )  
     else 
       let val s = compType (EndPoint (e), EndPoint ( Qualified (Un,r)))
        in let val t=downCast(s) 
         in Channel ( t, Qualified (Un,p))
        end
       end
 | compType (Channel   ( Qualified (Lin,p),  e), EndPoint  ( Qualified (Lin,r)))=
     if p=r   then 
       Channel ( Void,e )  
     else 
       let val s = compType (EndPoint (e), EndPoint ( Qualified (Lin,r)))
        in let val t=downCast(s) 
         in Channel ( Qualified (Lin,p),t )
        end
       end
  | compType (Channel   ( e,  Qualified (Lin,p)), EndPoint  ( Qualified (Lin,r)))=
     if p=r   then 
       Channel ( e,Void )  
     else 
       let val s = compType (EndPoint (e), EndPoint ( Qualified (Lin,r)))
        in let val t=downCast(s) 
         in Channel ( t, Qualified (Lin,p))
        end
       end
  | compType (_,_)= raise checkVarFail	;


(***************************Functions over contexts****************************) 

(*** Put on top of the stack a variable ***)

fun topStack (nil:context,v:var):context= raise contextFail
|   topStack ((x,s)::g,y) =
     if x=y then 
      (x,s)::g
     else
      let val d = topStack (g,y)
       in d@[(x,s)]
      end


(***Removing an entry from the context***)
fun remove (nil:context,v:var):context = raise contextFail
 |  remove ((x,t)::g,v)  =
     if x=v then g
     else 
      let val d = remove(g,v)
       in (x,t)::d
      end;  

     
(*** Checking if the the type associated to a variable is
unrestricted or void. Used in rules for input and output in fun check***) 
fun consumed (nil:context, v:var):context =  raise contextFail
 |  consumed ((x,s)::g,v) = 
     if x=v then 
      let val t= unRestricted (s) 
       in
         (x,t)::g 
      end
     else
      let val d = consumed (g,v)
       in (x,s)::d
      end;    
      
(*** Checking if the environment has been all consumed ***)
fun finished (nil)=nil
 |  finished ((x,s)::g) = 
     let val t= unRestricted (s) 
       in (x,t):: finished (g)
     end

(*** Setting a end point type in the domain of a context to void ***)
fun setVoid (nil:context,v:var):context = raise contextFail
 |  setVoid ((x,EndPoint s)::g,v:var) =
      if x=v then 
       (x,EndPoint Void)::g
      else
       let val d = setVoid(g,v)
        in (x,EndPoint s)::d
       end
 |  setVoid (_,_) = raise typeFail;

(*** Transforming an end point type in the domain of a context in a channel type ***)
fun upCastContext (nil:context,v:var, EndPoint e):context = raise contextFail
 |  upCastContext ((x,EndPoint h)::g, v, EndPoint e) =
     if x=v then
      let val s = upCast (EndPoint h,EndPoint e) 
       in (x,s)::g
      end
     else
      let val d = topStack ( (x,EndPoint h)::g,v)
       in  upCastContext (d,v,EndPoint e)
      end
 | upCastContext(_,_,_)= raise contextFail;

fun safeContext (nil)=nil
 |  safeContext ((x,s)::g) =
      (x,safe(s))::safeContext (g); 
 
(**************************Type Checking************************************)

(*** Type checking a variable. Used in the patterns for output in fun check***)

fun checkVar( nil:context, (z,t):(var * sessionType) ):context = raise checkVarFail 
 |  checkVar( (x,s)::g, (z,t) ) = 
     if x=z then
      let val u = compType(s,t) in
       (x,u)::g
      end
     else
      let val d=checkVar (g,(z,t)) in
        (x,s)::d
      end;
 
(***************************** Type checking processes *****************************)

(* Casting a string in output *)
fun stringCast (Var (s))=s
 |  stringCast(_) = raise typeFail;

(* Checking the inert process*)
(* Any context suffices *)
(* A-INACT *)
fun check (g:context,Zero)  = g               

(* Checking the replicated process*)
(* Since only linear types evolve, to forbid the use of a linear resource under
replication we require that the returned context is equal to the input context *)
(* A-REPL  *)
 |  check (g:context,Replication p):context     =                  
    if g = check(g,p) 
      then g
    else raise checkFail
    
(* Checking a parallel composition***)    
(* Elegant and compact, improves current technology for linear and session type checking.
The wedge of the formulation is that the domain of the returned context is equal to
the domain of the input context. $Void$ types account for  linear resources already
consumed *)
(* The order of the processes is irrelevant for the top level call. For a formal proof of
this claim see [1] *)
(* A-PAR   *)
 |  check (g:context, Parallel (p1,p2)):context =                 
      check( check(g,p1),p2 )                   

(* Checking the input process***)

(* Case whether the type for the input is an unrestricted end point. *)
(* The call for the continuation is invoked by passing the variable $y$ bound by the input
at the argument type $a$,  and  the input context:  unrestricted types do not
evolve.  *)
(*A-IN-UN *)
|  check ( (z,(EndPoint  ( Qualified (Un,In(a,c)) )))::g , Input (x,y,p) ) =
      if x=z then
       let val d = check ( (y,a)::(z,(EndPoint  ( Qualified (Un,In(a,c)) )))::g , p)  
        in  remove ( consumed (d,y), y)
       end
      else
       let val d = topStack ( (z,(EndPoint  ( Qualified (Un,In(a,c)) )))::g, x)
        in check ( d , Input (x,y,p) )
       end
(* Case whether the type for the input is a linear end point. *)
(* The call for the continuation is invoked by passing the variable $y$ bound by the input
at the argument type $a$, the continuation type $c$ for $x$, and the tail of the context: 
linear types do evolve. *)
(*A-IN-L *)
|  check ( (z,(EndPoint  ( Qualified (Lin,In(a,c)) )))::g , Input (x,y,p) ) =
      if x=z then
       let val d = check ( (y,a)::(x,EndPoint (c))::g,p) 
        in setVoid ( remove ( consumed (consumed (d,x), y), y) , x)
       end
      else
       let val d = topStack ( (z,(EndPoint  ( Qualified (Lin,In(a,c)) )))::g, x)
        in check ( d , Input (x,y,p) )
       end
(* Cases whether the type for the input is an unrestricted entry in a channel type. 
Implemented by the case for the unrestricted end point A-IN-UN. *)

(*A-IN-UN-l *)
|  check ( (z,Channel  ( Qualified (Un,In(a,c)) ,e ))::g , Input (x,y,p) ) = 
      if x=z then
       let val d = check ( (y,a)::(z,Channel  ( Qualified (Un,In(a,c)) ,e ))::g , p)  
        in  remove ( consumed (d,y), y)
       end
      else
       let val d = topStack ( (z,Channel  ( Qualified (Un,In(a,c)) ,e ))::g, x)
        in check ( d , Input (x,y,p) )
       end
(*A-IN-UN-r *)
|  check ( (z,Channel  ( e, Qualified (Un,In(a,c))))::g , Input (x,y,p) ) = 
      if x=z then
       let val d = check ( (y,a)::(z,Channel  ( e, Qualified (Un,In(a,c))))::g , p)  
        in  remove ( consumed (d,y), y)
       end
      else
       let val d = topStack ( (z,Channel  ( e, Qualified (Un,In(a,c))))::g, x)
        in check ( d , Input (x,y,p) )
       end
(* Cases whether the type for the input is a linear entry in a channel type. 
Implemented by the case for the linear end point A-IN-L. *)

(*A-IN-L-l *)
|  check ( (z,(Channel  ( Qualified (Lin,In(a,c)) , e)))::g , Input (x,y,p) ) =
      if x=z then
       let val d = check ((z,(EndPoint  ( Qualified (Lin,In(a,c)))))::g , Input (x,y,p)) 
        in upCastContext ( d, x, EndPoint e)
       end
      else
       let val d = topStack ( (z,(EndPoint  ( Qualified (Lin,In(a,c)) )))::g, x)
        in check ( d , Input (x,y,p) )
       end
(*A-IN-L-r *)
|  check ( (z,(Channel  ( e, Qualified (Lin,In(a,c)))))::g , Input (x,y,p) ) =
      if x=z then
       let val d = check ((z,(EndPoint  ( Qualified (Lin,In(a,c)))))::g , Input (x,y,p)) 
        in upCastContext ( d, x, EndPoint e)
       end
      else
       let val d = topStack ( (z,(EndPoint  ( Qualified (Lin,In(a,c)) )))::g, x)
        in check ( d , Input (x,y,p) )
       end
(* Input - Else *)
|  check ( (z,s)::g, Input (x,y,p)) = 
     if x=z then raise checkFail
     else
      let val d = topStack ((z,s)::g, x)
       in check ( d, Input (x,y,p))
      end 
      
      
(* Checking the output process *)

(*  Case whether the type for the output is an unrestricted end point. *)
(*  We first check with the input context the variable $y$ to be sent . We then use the
context $h$ received from the $checkVar$ call to control the continuation. If $a$ is a
linear type, then its type in $h$ is Void. This implements session delegation. *)
(*A-OUT-UN *)
|  check ( (z,(EndPoint  ( Qualified (Un,Out(a,c)) )))::g , Output (x,Var (y),p) ) =
      if x=z then
       let val h = checkVar ((z,(EndPoint  ( Qualified (Un,Out(a,c)) )))::g,
                             (stringCast (Var (y)),a)) 
         in check ( (z,(EndPoint  ( Qualified (Un,Out(a,c)))))::h ,p)  
       end
      else
       let val d = topStack ( (z,(EndPoint  ( Qualified (Un,Out(a,c)) )))::g, x)
        in check ( d , Output (x,Var (y),p) )
       end
(* When a value is sent we skip the call to checkVar and invokes the call for the
continuation. The value is the parameter given to a $process$ datatype *)
| check ( (z,(EndPoint  ( Qualified (Un,Out(Base,c)) )))::g , Output (x,Value (y),p) ) =
      if x=z then 
        check ( (z,(EndPoint  ( Qualified (Un,Out(Base,c)) )))::g ,p)   
      else
       let val d = topStack ( (z,(EndPoint  ( Qualified (Un,Out(Base,c)) )))::g, x)
        in check ( d , Output (x,Value (y),p) )
       end
(*  Case whether the type for the output is a linear end point. *)
(*A-OUT-L *)
|  check ( (z,(EndPoint  ( Qualified (Lin,Out(a,c)) )))::g , Output (x,Var (y),p) ) =
      if x=z then
       let val h = checkVar (g,(stringCast (Var (y)),a)) in 
         let val d = check ( (z,EndPoint (c))::h ,p) in 
           consumed (d,z)
         end 
       end
      else
       let val d = topStack ( (z,(EndPoint  ( Qualified (Lin,Out(a,c)) )))::g, x)
        in check ( d , Output (x,Var (y),p) )
       end 
|  check ( (z,(EndPoint  ( Qualified (Lin,Out(Base,c)) )))::g , Output (x,Value (y),p) ) =
      if x=z then 
         let val d = check ( (z,EndPoint (c))::g ,p) in 
           consumed (d,z)
         end 
      else
       let val d = topStack ( (z,(EndPoint  ( Qualified (Lin,Out(Base,c)) )))::g, x)
        in check ( d , Output (x,Value (y),p) )
       end
(* Cases whether the type for the output is an unrestricted entry in a channel type. 
Implemented by the case for the unrestricted end point A-OUT-UN *)

(*A-OUT-UN-l *)
|  check ( (z,(Channel ( Qualified (Un,Out(a,c)) ,e)))::g , Output (x,Var (y),p) )=
      if x=z then
       let val h = checkVar ((z,(Channel ( Qualified (Un,Out(a,c)) ,e)))::g,
                             (stringCast (Var (y)),a)) 
         in check ( h ,p)  
       end
      else
       let val d = topStack ( (z,(Channel ( Qualified (Un,Out(a,c)) ,e)))::g, x)
        in check ( d , Output (x,Var (y),p) )
       end
|  check ( (z,(Channel ( Qualified (Un,Out(Base,c)) ,e)))::g , Output (x,Value (y),p) )=
      if x=z then   
          check ( (z,(Channel ( Qualified (Un,Out(Base,c)) ,e)))::g ,p)   
      else
       let val d = topStack ( (z,(Channel ( Qualified (Un,Out(Base,c)) ,e)))::g, x)
        in check ( d , Output (x,Value (y),p) )
       end
(*A-OUT-UN-r *)
|  check ( (z,(Channel ( e, Qualified (Un,Out(a,c)))))::g , Output (x,Var (y),p) )=
      if x=z then
       let val h = checkVar (g,(stringCast (Var (y)),a)) 
         in check ( (z,(Channel ( e, Qualified (Un,Out(a,c)))))::h ,p)  
       end
      else
       let val d = topStack ( (z,(Channel ( e, Qualified (Un,Out(a,c)))))::g, x)
        in check ( d , Output (x,Var (y),p) )
       end
(* Cases whether the type for the output is an unrestricted entry in a channel type. 
Implemented by the case for the unrestricted end point A-OUT-L *)

(*A-OUT-L-l *)
|  check ( (z,(Channel ( Qualified (Lin,Out(a,c)) ,e)))::g , Output (x,y,p) )=
      if x=z then
       let val d = check ((z,(EndPoint ( Qualified (Lin,Out(a,c)))))::g , 
                          Output (x,y,p))
       in  upCastContext ( d, x, EndPoint e)
       end
      else
       let val d = topStack ( (z,(Channel ( Qualified (Lin,Out(a,c)) ,e)))::g, x)
        in check ( d , Output (x,y ,p) )
       end
(*A-OUT-L-r *)
|  check ( (z,(Channel ( e, Qualified (Lin,Out(a,c)))))::g , Output (x,y,p) )=
      if x=z then
       let val d = check ((z,(EndPoint ( Qualified (Lin,Out(a,c)))))::g , 
                          Output (x,y,p))
       in  upCastContext ( d, x, EndPoint e)
       end
      else
       let val d = topStack ( (z,(Channel ( e, Qualified (Lin,Out(a,c)))))::g, x)
        in check ( d , Output (x,y,p) )
       end      
(* Output - Else *)
|  check ( (z,s)::g, Output (x,y,p)) = 
     if x=z then raise checkFail
     else
      let val d = topStack ((z,s)::g, x)
       in check ( d, Output (x,y,p))
      end

(* Checking the restriction process *)
(* To preserve subject reduction (cf. [2]) the type of the new channel need to be safe or
the two channel ends of the session must exchange values of the same type. *)

| check (g, New (x,t,p) )  =    
     remove ( consumed ( check ( (x,safe(t))::g, p) , x ) , x)

| check (_,_) = raise checkFail;
     
(******************************* Top level call****************************************)
(* Requires a safe context to preserve typability during the computation, or the subject
reduction property: see above. When the call does not raise an exception, a
session based type discipline is established, and preserved during the computation. The
functions does return the context consumed by the process, i.e. the context
with all linear types in the domain of the  input context set to $Void$. *)


fun typeCheck (g,p)  =   
      finished ( check ( safeContext(g) ,p ) );


(***************************** References *******************************

[1] Marco Giunti. A type checking algorithm for qualified session types. Workshop on
Automated Specification and Verification of Web Systems (WWV), 2011. To appear in
Electronic Proceedings in Theoretical Computer Science
[2] Marco Giunti, Vasco T. Vasconcelos: A Linear Account of Session Types in the
Pi Calculus. Concurrency Theory, 21th International Conference (CONCUR). Lecture Notes in
Computer Science 6269, 432-446, 2010.
[3] Kohei Honda, Vasco T. Vasconcelos, Makoto Kubo: Language Primitives and Type
Discipline for Structured Communication-Based Programming. Programming Languages and
Systems (ESOP).Lecture Notes in Computer Science 1381, 122-138, 1998.
[4] Robin Milner, Joachim Parrow, David Walker: A Calculus of Mobile Processes, I and II.
Information and Computation, 100(1): 1-77, 1992.
[5] A Conversation with Steve Ross-Talbot. ACM Queue, 2006.
http://queue.acm.org/detail.cfm?id=1122684

*)

 